@@ -1,6 +1,6 @@
 /*
- * SSLsplit - transparent and scalable SSL/TLS interception
- * Copyright (c) 2009-2014, Daniel Roethlisberger <daniel@roe.ch>
+ * SSLsplit - transparent SSL/TLS interception
+ * Copyright (c) 2009-2016, Daniel Roethlisberger <daniel@roe.ch>
  * All rights reserved.
  * http://www.roe.ch/SSLsplit
  *
@@ -8,8 +8,7 @@
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
- *    notice unmodified, this list of conditions, and the following
- *    disclaimer.
+ *    notice, this list of conditions, and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
@@ -120,6 +119,7 @@ typedef struct pxy_conn_ctx {
 
 	/* status flags */
 	unsigned int immutable_cert : 1;  /* 1 if the cert cannot be changed */
+	unsigned int generated_cert : 1;     /* 1 if we generated a new cert */
 	unsigned int connected : 1;       /* 0 until both ends are connected */
 	unsigned int seen_req_header : 1; /* 0 until request header complete */
 	unsigned int seen_resp_header : 1;  /* 0 until response hdr complete */
@@ -128,13 +128,17 @@ typedef struct pxy_conn_ctx {
 	unsigned int ocsp_denied : 1;                /* 1 if OCSP was denied */
 	unsigned int enomem : 1;                       /* 1 if out of memory */
 	unsigned int sni_peek_retries : 6;       /* max 64 SNI parse retries */
+	unsigned int clienthello_search : 1;       /* 1 if waiting for hello */
+	unsigned int clienthello_found : 1;      /* 1 if conn upgrade to SSL */
 
 	/* server name indicated by client in SNI TLS extension */
 	char *sni;
 
 	/* log strings from socket */
-	char *src_str;
-	char *dst_str;
+	char *srchost_str;
+	char *srcport_str;
+	char *dsthost_str;
+	char *dstport_str;
 
 	/* log strings from HTTP request */
 	char *http_method;
@@ -147,8 +151,10 @@ typedef struct pxy_conn_ctx {
 	char *http_status_text;
 	char *http_content_length;
 
-	/* log strings from SSL context */
+	/* log strings related to SSL */
 	char *ssl_names;
+	char *origcrtfpr;
+	char *usedcrtfpr;
 
 #ifdef HAVE_LOCAL_PROCINFO
 	/* local process information */
@@ -194,6 +200,7 @@ pxy_conn_ctx_new(proxyspec_t *spec, opts_t *opts,
 	memset(ctx, 0, sizeof(pxy_conn_ctx_t));
 	ctx->spec = spec;
 	ctx->opts = opts;
+	ctx->clienthello_search = spec->upgrade;
 	ctx->fd = fd;
 	ctx->thridx = pxy_thrmgr_attach(thrmgr, &ctx->evbase, &ctx->dnsbase);
 	ctx->thrmgr = thrmgr;
@@ -209,9 +216,7 @@ pxy_conn_ctx_new(proxyspec_t *spec, opts_t *opts,
 	return ctx;
 }
 
-static void
-pxy_conn_ctx_free(pxy_conn_ctx_t *ctx) NONNULL(1);
-static void
+static void NONNULL(1)
 pxy_conn_ctx_free(pxy_conn_ctx_t *ctx)
 {
 #ifdef DEBUG_PROXY
@@ -221,11 +226,17 @@ pxy_conn_ctx_free(pxy_conn_ctx_t *ctx)
 	}
 #endif /* DEBUG_PROXY */
 	pxy_thrmgr_detach(ctx->thrmgr, ctx->thridx);
-	if (ctx->src_str) {
-		free(ctx->src_str);
+	if (ctx->srchost_str) {
+		free(ctx->srchost_str);
 	}
-	if (ctx->dst_str) {
-		free(ctx->dst_str);
+	if (ctx->srcport_str) {
+		free(ctx->srcport_str);
+	}
+	if (ctx->dsthost_str) {
+		free(ctx->dsthost_str);
+	}
+	if (ctx->dstport_str) {
+		free(ctx->dstport_str);
 	}
 	if (ctx->http_method) {
 		free(ctx->http_method);
@@ -251,6 +262,12 @@ pxy_conn_ctx_free(pxy_conn_ctx_t *ctx)
 	if (ctx->ssl_names) {
 		free(ctx->ssl_names);
 	}
+	if (ctx->origcrtfpr) {
+		free(ctx->origcrtfpr);
+	}
+	if (ctx->usedcrtfpr) {
+		free(ctx->usedcrtfpr);
+	}
 #ifdef HAVE_LOCAL_PROCINFO
 	if (ctx->lproc.exec_path) {
 		free(ctx->lproc.exec_path);
@@ -312,17 +329,12 @@ pxy_debug_crt(X509 *crt)
 		free(names);
 	}
 
-	unsigned char fpr[SSL_X509_FPRSZ];
-	if (ssl_x509_fingerprint_sha1(crt, fpr) == -1) {
+	char *fpr;
+	if (!(fpr = ssl_x509_fingerprint(crt, 1))) {
 		log_err_printf("Warning: Error generating X509 fingerprint\n");
 	} else {
-		log_dbg_printf("Fingerprint: "     "%02x:%02x:%02x:%02x:"
-		               "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:"
-		               "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x\n",
-		               fpr[0],  fpr[1],  fpr[2],  fpr[3],  fpr[4],
-		               fpr[5],  fpr[6],  fpr[7],  fpr[8],  fpr[9],
-		               fpr[10], fpr[11], fpr[12], fpr[13], fpr[14],
-		               fpr[15], fpr[16], fpr[17], fpr[18], fpr[19]);
+		log_dbg_printf("Fingerprint: %s\n", fpr);
+		free(fpr);
 	}
 
 #ifdef DEBUG_CERTIFICATE
@@ -357,39 +369,47 @@ pxy_log_connect_nonhttp(pxy_conn_ctx_t *ctx)
 	}
 #endif /* HAVE_LOCAL_PROCINFO */
 
-	if (!ctx->spec->ssl || ctx->passthrough) {
-		rv = asprintf(&msg, "%s %s %s"
+	if (!ctx->src.ssl) {
+		rv = asprintf(&msg, "%s %s %s %s %s"
 #ifdef HAVE_LOCAL_PROCINFO
 		              " %s"
 #endif /* HAVE_LOCAL_PROCINFO */
 		              "\n",
 		              ctx->passthrough ? "passthrough" : "tcp",
-		              STRORDASH(ctx->src_str),
-		              STRORDASH(ctx->dst_str)
+		              STRORDASH(ctx->srchost_str),
+		              STRORDASH(ctx->srcport_str),
+		              STRORDASH(ctx->dsthost_str),
+		              STRORDASH(ctx->dstport_str)
 #ifdef HAVE_LOCAL_PROCINFO
 		              , lpi
 #endif /* HAVE_LOCAL_PROCINFO */
 		             );
 	} else {
-		rv = asprintf(&msg, "ssl %s %s "
+		rv = asprintf(&msg, "%s %s %s %s %s "
 		              "sni:%s names:%s "
-		              "sproto:%s:%s dproto:%s:%s"
+		              "sproto:%s:%s dproto:%s:%s "
+		              "origcrt:%s usedcrt:%s"
 #ifdef HAVE_LOCAL_PROCINFO
 		              " %s"
 #endif /* HAVE_LOCAL_PROCINFO */
 		              "\n",
-		              STRORDASH(ctx->src_str),
-		              STRORDASH(ctx->dst_str),
+		              ctx->clienthello_found ? "upgrade" : "ssl",
+		              STRORDASH(ctx->srchost_str),
+		              STRORDASH(ctx->srcport_str),
+		              STRORDASH(ctx->dsthost_str),
+		              STRORDASH(ctx->dstport_str),
 		              STRORDASH(ctx->sni),
 		              STRORDASH(ctx->ssl_names),
 		              SSL_get_version(ctx->src.ssl),
 		              SSL_get_cipher(ctx->src.ssl),
 		              SSL_get_version(ctx->dst.ssl),
-		              SSL_get_cipher(ctx->dst.ssl)
+		              SSL_get_cipher(ctx->dst.ssl),
+		              STRORDASH(ctx->origcrtfpr),
+		              STRORDASH(ctx->usedcrtfpr)
 #ifdef HAVE_LOCAL_PROCINFO
 		              , lpi
 #endif /* HAVE_LOCAL_PROCINFO */
-		             );
+		              );
 	}
 	if ((rv < 0) || !msg) {
 		ctx->enomem = 1;
@@ -447,13 +467,15 @@ pxy_log_connect_http(pxy_conn_ctx_t *ctx)
 #endif /* HAVE_LOCAL_PROCINFO */
 
 	if (!ctx->spec->ssl) {
-		rv = asprintf(&msg, "http %s %s %s %s %s %s %s"
+		rv = asprintf(&msg, "http %s %s %s %s %s %s %s %s %s"
 #ifdef HAVE_LOCAL_PROCINFO
 		              " %s"
 #endif /* HAVE_LOCAL_PROCINFO */
 		              "%s\n",
-		              STRORDASH(ctx->src_str),
-		              STRORDASH(ctx->dst_str),
+		              STRORDASH(ctx->srchost_str),
+		              STRORDASH(ctx->srcport_str),
+		              STRORDASH(ctx->dsthost_str),
+		              STRORDASH(ctx->dstport_str),
 		              STRORDASH(ctx->http_host),
 		              STRORDASH(ctx->http_method),
 		              STRORDASH(ctx->http_uri),
@@ -464,15 +486,18 @@ pxy_log_connect_http(pxy_conn_ctx_t *ctx)
 #endif /* HAVE_LOCAL_PROCINFO */
 		              ctx->ocsp_denied ? " ocsp:denied" : "");
 	} else {
-		rv = asprintf(&msg, "https %s %s %s %s %s %s %s "
+		rv = asprintf(&msg, "https %s %s %s %s %s %s %s %s %s "
 		              "sni:%s names:%s "
-		              "sproto:%s:%s dproto:%s:%s"
+		              "sproto:%s:%s dproto:%s:%s "
+		              "origcrt:%s usedcrt:%s"
 #ifdef HAVE_LOCAL_PROCINFO
 		              " %s"
 #endif /* HAVE_LOCAL_PROCINFO */
 		              "%s\n",
-		              STRORDASH(ctx->src_str),
-		              STRORDASH(ctx->dst_str),
+		              STRORDASH(ctx->srchost_str),
+		              STRORDASH(ctx->srcport_str),
+		              STRORDASH(ctx->dsthost_str),
+		              STRORDASH(ctx->dstport_str),
 		              STRORDASH(ctx->http_host),
 		              STRORDASH(ctx->http_method),
 		              STRORDASH(ctx->http_uri),
@@ -484,6 +509,8 @@ pxy_log_connect_http(pxy_conn_ctx_t *ctx)
 		              SSL_get_cipher(ctx->src.ssl),
 		              SSL_get_version(ctx->dst.ssl),
 		              SSL_get_cipher(ctx->dst.ssl),
+		              STRORDASH(ctx->origcrtfpr),
+		              STRORDASH(ctx->usedcrtfpr),
 #ifdef HAVE_LOCAL_PROCINFO
 		              lpi,
 #endif /* HAVE_LOCAL_PROCINFO */
@@ -520,11 +547,11 @@ out:
  * the refcount decrementing.  In other words, return 0 if we did not
  * keep a pointer to the object (which we never do here).
  */
-#ifdef WITH_SSLV2
+#ifdef HAVE_SSLV2
 #define MAYBE_UNUSED 
-#else /* !WITH_SSLV2 */
+#else /* !HAVE_SSLV2 */
 #define MAYBE_UNUSED UNUSED
-#endif /* !WITH_SSLV2 */
+#endif /* !HAVE_SSLV2 */
 static int
 pxy_ossl_sessnew_cb(MAYBE_UNUSED SSL *ssl, SSL_SESSION *sess)
 #undef MAYBE_UNUSED
@@ -534,10 +561,10 @@ pxy_ossl_sessnew_cb(MAYBE_UNUSED SSL *ssl, SSL_SESSION *sess)
 	if (sess) {
 		log_dbg_print_free(ssl_session_to_str(sess));
 	} else {
-		log_dbg_print("(null)\n");
+		log_dbg_printf("(null)\n");
 	}
 #endif /* DEBUG_SESSION_CACHE */
-#ifdef WITH_SSLV2
+#ifdef HAVE_SSLV2
 	/* Session resumption seems to fail for SSLv2 with protocol
 	 * parsing errors, so we disable caching for SSLv2. */
 	if (SSL_version(ssl) == SSL2_VERSION) {
@@ -545,7 +572,7 @@ pxy_ossl_sessnew_cb(MAYBE_UNUSED SSL *ssl, SSL_SESSION *sess)
 		               "client.\n");
 		return 0;
 	}
-#endif /* WITH_SSLV2 */
+#endif /* HAVE_SSLV2 */
 	if (sess) {
 		cachemgr_ssess_set(sess);
 	}
@@ -565,7 +592,7 @@ pxy_ossl_sessremove_cb(UNUSED SSL_CTX *sslctx, SSL_SESSION *sess)
 	if (sess) {
 		log_dbg_print_free(ssl_session_to_str(sess));
 	} else {
-		log_dbg_print("(null)\n");
+		log_dbg_printf("(null)\n");
 	}
 #endif /* DEBUG_SESSION_CACHE */
 	if (sess) {
@@ -599,16 +626,11 @@ pxy_ossl_sessget_cb(UNUSED SSL *ssl, unsigned char *id, int idlen, int *copy)
 }
 
 /*
- * Create and set up a new SSL_CTX instance for terminating SSL.
- * Set up all the necessary callbacks, the certificate, the cert chain and key.
+ * Set SSL_CTX options that are the same for incoming and outgoing SSL_CTX.
  */
-static SSL_CTX *
-pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
-                     EVP_PKEY *key)
+static void
+pxy_sslctx_setoptions(SSL_CTX *sslctx, pxy_conn_ctx_t *ctx)
 {
-	SSL_CTX *sslctx = SSL_CTX_new(ctx->opts->sslmethod());
-	if (!sslctx)
-		return NULL;
 	SSL_CTX_set_options(sslctx, SSL_OP_ALL);
 #ifdef SSL_OP_TLS_ROLLBACK_BUG
 	SSL_CTX_set_options(sslctx, SSL_OP_TLS_ROLLBACK_BUG);
@@ -622,16 +644,11 @@ pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
 #ifdef SSL_OP_NO_TICKET
 	SSL_CTX_set_options(sslctx, SSL_OP_NO_TICKET);
 #endif /* SSL_OP_NO_TICKET */
-#ifdef SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION
-	SSL_CTX_set_options(sslctx,
-	                    SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION);
-#endif /* SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION */
-#ifdef SSL_OP_NO_COMPRESSION
-	if (!ctx->opts->sslcomp) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_COMPRESSION);
-	}
-#endif /* SSL_OP_NO_COMPRESSION */
 
+	/*
+	 * Do not use HAVE_SSLV2 because we need to set SSL_OP_NO_SSLv2 if it
+	 * is available and WITH_SSLV2 was not used.
+	 */
 #ifdef SSL_OP_NO_SSLv2
 #ifdef WITH_SSLV2
 	if (ctx->opts->no_ssl2) {
@@ -641,28 +658,50 @@ pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
 	}
 #endif /* WITH_SSLV2 */
 #endif /* !SSL_OP_NO_SSLv2 */
-#ifdef SSL_OP_NO_SSLv3
+#ifdef HAVE_SSLV3
 	if (ctx->opts->no_ssl3) {
 		SSL_CTX_set_options(sslctx, SSL_OP_NO_SSLv3);
 	}
-#endif /* SSL_OP_NO_SSLv3 */
-#ifdef SSL_OP_NO_TLSv1
+#endif /* HAVE_SSLV3 */
+#ifdef HAVE_TLSV10
 	if (ctx->opts->no_tls10) {
 		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1);
 	}
-#endif /* SSL_OP_NO_TLSv1 */
-#ifdef SSL_OP_NO_TLSv1_1
+#endif /* HAVE_TLSV10 */
+#ifdef HAVE_TLSV11
 	if (ctx->opts->no_tls11) {
 		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1_1);
 	}
-#endif /* SSL_OP_NO_TLSv1_1 */
-#ifdef SSL_OP_NO_TLSv1_2
+#endif /* HAVE_TLSV11 */
+#ifdef HAVE_TLSV12
 	if (ctx->opts->no_tls12) {
 		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1_2);
 	}
-#endif /* SSL_OP_NO_TLSv1_2 */
+#endif /* HAVE_TLSV12 */
+
+#ifdef SSL_OP_NO_COMPRESSION
+	if (!ctx->opts->sslcomp) {
+		SSL_CTX_set_options(sslctx, SSL_OP_NO_COMPRESSION);
+	}
+#endif /* SSL_OP_NO_COMPRESSION */
 
 	SSL_CTX_set_cipher_list(sslctx, ctx->opts->ciphers);
+}
+
+/*
+ * Create and set up a new SSL_CTX instance for terminating SSL.
+ * Set up all the necessary callbacks, the certificate, the cert chain and key.
+ */
+static SSL_CTX *
+pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
+                     EVP_PKEY *key)
+{
+	SSL_CTX *sslctx = SSL_CTX_new(ctx->opts->sslmethod());
+	if (!sslctx)
+		return NULL;
+
+	pxy_sslctx_setoptions(sslctx, ctx);
+
 	SSL_CTX_sess_set_new_cb(sslctx, pxy_ossl_sessnew_cb);
 	SSL_CTX_sess_set_remove_cb(sslctx, pxy_ossl_sessremove_cb);
 	SSL_CTX_sess_set_get_cb(sslctx, pxy_ossl_sessget_cb);
@@ -679,7 +718,7 @@ pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
 #ifndef OPENSSL_NO_DH
 	if (ctx->opts->dh) {
 		SSL_CTX_set_tmp_dh(sslctx, ctx->opts->dh);
-	} else if (EVP_PKEY_type(key->type) != EVP_PKEY_RSA) {
+	} else {
 		SSL_CTX_set_tmp_dh_callback(sslctx, ssl_tmp_dh_callback);
 	}
 #endif /* !OPENSSL_NO_DH */
@@ -688,7 +727,7 @@ pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
 		EC_KEY *ecdh = ssl_ec_by_name(ctx->opts->ecdhcurve);
 		SSL_CTX_set_tmp_ecdh(sslctx, ecdh);
 		EC_KEY_free(ecdh);
-	} else if (EVP_PKEY_type(key->type) != EVP_PKEY_RSA) {
+	} else {
 		EC_KEY *ecdh = ssl_ec_by_name(NULL);
 		SSL_CTX_set_tmp_ecdh(sslctx, ecdh);
 		EC_KEY_free(ecdh);
@@ -724,6 +763,48 @@ pxy_srcsslctx_create(pxy_conn_ctx_t *ctx, X509 *crt, STACK_OF(X509) *chain,
 	return sslctx;
 }
 
+static int
+pxy_srccert_write_to_gendir(pxy_conn_ctx_t *ctx, X509 *crt, int is_orig)
+{
+	char *fn;
+	int rv;
+
+	if (!ctx->origcrtfpr)
+		return -1;
+	if (is_orig) {
+		rv = asprintf(&fn, "%s/%s.crt", ctx->opts->certgendir,
+		              ctx->origcrtfpr);
+	} else {
+		if (!ctx->usedcrtfpr)
+			return -1;
+		rv = asprintf(&fn, "%s/%s-%s.crt", ctx->opts->certgendir,
+		              ctx->origcrtfpr, ctx->usedcrtfpr);
+	}
+	if (rv == -1) {
+		ctx->enomem = 1;
+		return -1;
+	}
+	rv = log_cert_submit(fn, crt);
+	free(fn);
+	return rv;
+}
+
+static void
+pxy_srccert_write(pxy_conn_ctx_t *ctx)
+{
+	if (ctx->opts->certgen_writeall || ctx->generated_cert) {
+		if (pxy_srccert_write_to_gendir(ctx,
+		                SSL_get_certificate(ctx->src.ssl), 0) == -1) {
+			log_err_printf("Failed to write used certificate\n");
+		}
+	}
+	if (ctx->opts->certgen_writeall) {
+		if (pxy_srccert_write_to_gendir(ctx, ctx->origcrt, 1) == -1) {
+			log_err_printf("Failed to write orig certificate\n");
+		}
+	}
+}
+
 static cert_t *
 pxy_srccert_create(pxy_conn_ctx_t *ctx)
 {
@@ -796,6 +877,19 @@ pxy_srccert_create(pxy_conn_ctx_t *ctx)
 		}
 		cert_set_key(cert, ctx->opts->key);
 		cert_set_chain(cert, ctx->opts->chain);
+		ctx->generated_cert = 1;
+	}
+
+	if ((WANT_CONNECT_LOG(ctx) || ctx->opts->certgendir) && ctx->origcrt) {
+		ctx->origcrtfpr = ssl_x509_fingerprint(ctx->origcrt, 0);
+		if (!ctx->origcrtfpr)
+			ctx->enomem = 1;
+	}
+	if ((WANT_CONNECT_LOG(ctx) || ctx->opts->certgen_writeall) &&
+	    cert && cert->crt) {
+		ctx->usedcrtfpr = ssl_x509_fingerprint(cert->crt, 0);
+		if (!ctx->usedcrtfpr)
+			ctx->enomem = 1;
 	}
 
 	return cert;
@@ -883,6 +977,18 @@ pxy_ossl_servername_cb(SSL *ssl, UNUSED int *al, void *arg)
 	if (!(sn = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name)))
 		return SSL_TLSEXT_ERR_NOACK;
 
+	if (!ctx->sni) {
+		if (OPTS_DEBUG(ctx->opts)) {
+			log_dbg_printf("Warning: SNI parser yielded no "
+			               "hostname, copying OpenSSL one: "
+			               "[NULL] != [%s]\n", sn);
+		}
+		ctx->sni = strdup(sn);
+		if (!ctx->sni) {
+			ctx->enomem = 1;
+			return SSL_TLSEXT_ERR_NOACK;
+		}
+	}
 	if (OPTS_DEBUG(ctx->opts)) {
 		if (!!strcmp(sn, ctx->sni)) {
 			/*
@@ -894,7 +1000,7 @@ pxy_ossl_servername_cb(SSL *ssl, UNUSED int *al, void *arg)
 			 * to the original destination, there is no way back.
 			 * We log an error and hope this never happens.
 			 */
-			log_err_printf("Warning: SNI parser yielded different "
+			log_dbg_printf("Warning: SNI parser yielded different "
 			               "hostname than OpenSSL callback for "
 			               "the same ClientHello message: "
 			               "[%s] != [%s]\n", ctx->sni, sn);
@@ -919,6 +1025,7 @@ pxy_ossl_servername_cb(SSL *ssl, UNUSED int *al, void *arg)
 			return SSL_TLSEXT_ERR_NOACK;
 		}
 		cachemgr_fkcrt_set(ctx->origcrt, newcrt);
+		ctx->generated_cert = 1;
 		if (OPTS_DEBUG(ctx->opts)) {
 			log_dbg_printf("===> Updated forged server "
 			               "certificate:\n");
@@ -933,6 +1040,16 @@ pxy_ossl_servername_cb(SSL *ssl, UNUSED int *al, void *arg)
 				ctx->enomem = 1;
 			}
 		}
+		if (WANT_CONNECT_LOG(ctx) || ctx->opts->certgendir) {
+			if (ctx->usedcrtfpr) {
+				free(ctx->usedcrtfpr);
+			}
+			ctx->usedcrtfpr = ssl_x509_fingerprint(newcrt, 0);
+			if (!ctx->usedcrtfpr) {
+				ctx->enomem = 1;
+			}
+		}
+
 		newsslctx = pxy_srcsslctx_create(ctx, newcrt, ctx->opts->chain,
 		                                 ctx->opts->key);
 		if (!newsslctx) {
@@ -969,56 +1086,8 @@ pxy_dstssl_create(pxy_conn_ctx_t *ctx)
 		return NULL;
 	}
 
-	SSL_CTX_set_options(sslctx, SSL_OP_ALL);
-#ifdef SSL_OP_TLS_ROLLBACK_BUG
-	SSL_CTX_set_options(sslctx, SSL_OP_TLS_ROLLBACK_BUG);
-#endif /* SSL_OP_TLS_ROLLBACK_BUG */
-#ifdef SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
-	SSL_CTX_set_options(sslctx, SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION);
-#endif /* SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION */
-#ifdef SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
-	SSL_CTX_set_options(sslctx, SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS);
-#endif /* SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS */
-#ifdef SSL_OP_NO_TICKET
-	SSL_CTX_set_options(sslctx, SSL_OP_NO_TICKET);
-#endif /* SSL_OP_NO_TICKET */
-#ifdef SSL_OP_NO_COMPRESSION
-	if (!ctx->opts->sslcomp) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_COMPRESSION);
-	}
-#endif /* SSL_OP_NO_COMPRESSION */
+	pxy_sslctx_setoptions(sslctx, ctx);
 
-#ifdef SSL_OP_NO_SSLv2
-#ifdef WITH_SSLV2
-	if (ctx->opts->no_ssl2) {
-#endif /* WITH_SSLV2 */
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_SSLv2);
-#ifdef WITH_SSLV2
-	}
-#endif /* WITH_SSLV2 */
-#endif /* !SSL_OP_NO_SSLv2 */
-#ifdef SSL_OP_NO_SSLv3
-	if (ctx->opts->no_ssl3) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_SSLv3);
-	}
-#endif /* SSL_OP_NO_SSLv3 */
-#ifdef SSL_OP_NO_TLSv1
-	if (ctx->opts->no_tls10) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1);
-	}
-#endif /* SSL_OP_NO_TLSv1 */
-#ifdef SSL_OP_NO_TLSv1_1
-	if (ctx->opts->no_tls11) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1_1);
-	}
-#endif /* SSL_OP_NO_TLSv1_1 */
-#ifdef SSL_OP_NO_TLSv1_2
-	if (ctx->opts->no_tls12) {
-		SSL_CTX_set_options(sslctx, SSL_OP_NO_TLSv1_2);
-	}
-#endif /* SSL_OP_NO_TLSv1_2 */
-
-	SSL_CTX_set_cipher_list(sslctx, ctx->opts->ciphers);
 	SSL_CTX_set_verify(sslctx, SSL_VERIFY_NONE, NULL);
 
 	ssl = SSL_new(sslctx);
@@ -1399,6 +1468,7 @@ deny:
 		evbuffer_drain(inbuf, evbuffer_get_length(inbuf));
 	}
 	bufferevent_free_and_close_fd(ctx->dst.bev, ctx);
+	ctx->dst.bev = NULL;
 	ctx->dst.closed = 1;
 	evbuffer_add_printf(outbuf, ocspresp);
 	ctx->ocsp_denied = 1;
@@ -1417,6 +1487,71 @@ deny:
 	}
 }
 
+/*
+ * Peek into pending data to see if it is an SSL/TLS ClientHello, and if so,
+ * upgrade the connection from plain TCP to SSL/TLS.
+ *
+ * Return 1 if ClientHello was found and connection was upgraded to SSL/TLS,
+ * 0 otherwise.
+ *
+ * WARNING: This is experimental code and will need to be improved.
+ *
+ * TODO - enable search and skip bytes before ClientHello in case it does not
+ *        start at offset 0 (i.e. chello > vec_out[0].iov_base)
+ * TODO - peek into more than just the current segment
+ * TODO - add retry mechanism for short truncated ClientHello, possibly generic
+ */
+int
+pxy_conn_autossl_peek_and_upgrade(pxy_conn_ctx_t *ctx)
+{
+	struct evbuffer *inbuf;
+	struct evbuffer_iovec vec_out[1];
+	const unsigned char *chello;
+	if (OPTS_DEBUG(ctx->opts)) {
+		log_dbg_printf("Checking for a client hello\n");
+	}
+	/* peek the buffer */
+	inbuf = bufferevent_get_input(ctx->src.bev);
+	if (evbuffer_peek(inbuf, 1024, 0, vec_out, 1)) {
+		if (ssl_tls_clienthello_parse(vec_out[0].iov_base,
+		                              vec_out[0].iov_len,
+		                              0, &chello, &ctx->sni) == 0) {
+			if (OPTS_DEBUG(ctx->opts)) {
+				log_dbg_printf("Peek found ClientHello\n");
+			}
+			ctx->dst.ssl = pxy_dstssl_create(ctx);
+			if (!ctx->dst.ssl) {
+				log_err_printf("Error creating SSL for "
+				               "upgrade\n");
+				return 0;
+			}
+			ctx->dst.bev = bufferevent_openssl_filter_new(
+			               ctx->evbase, ctx->dst.bev, ctx->dst.ssl,
+			               BUFFEREVENT_SSL_CONNECTING, 0);
+			bufferevent_setcb(ctx->dst.bev, pxy_bev_readcb,
+			                  pxy_bev_writecb, pxy_bev_eventcb,
+			                  ctx);
+			bufferevent_enable(ctx->dst.bev, EV_READ|EV_WRITE);
+			if(!ctx->dst.bev) {
+				return 0;
+			}
+			if( OPTS_DEBUG(ctx->opts)) {
+				log_err_printf("Replaced dst bufferevent, new "
+				               "one is %p\n", ctx->dst.bev);
+			}
+			ctx->clienthello_search = 0;
+			ctx->clienthello_found = 1;
+			return 1;
+		} else {
+			if (OPTS_DEBUG(ctx->opts)) {
+				log_dbg_printf("Peek found no ClientHello\n");
+			}
+			return 0;
+		}
+	}
+	return 0;
+}
+
 void
 pxy_conn_terminate_free(pxy_conn_ctx_t *ctx)
 {
@@ -1424,15 +1559,17 @@ pxy_conn_terminate_free(pxy_conn_ctx_t *ctx)
 	               ctx->enomem ? " (out of memory)" : "");
 	if (ctx->dst.bev && !ctx->dst.closed) {
 		bufferevent_free_and_close_fd(ctx->dst.bev, ctx);
+		ctx->dst.bev = NULL;
 	}
 	if (ctx->src.bev && !ctx->src.closed) {
 		bufferevent_free_and_close_fd(ctx->src.bev, ctx);
+		ctx->src.bev = NULL;
 	}
 	pxy_conn_ctx_free(ctx);
 }
 
 /*
- * Callback for read events on the up- and downstram connection bufferevents.
+ * Callback for read events on the up- and downstream connection bufferevents.
  * Called when there is data ready in the input evbuffer.
  */
 static void
@@ -1456,6 +1593,12 @@ pxy_bev_readcb(struct bufferevent *bev, void *arg)
 		exit(EXIT_FAILURE);
 	}
 
+	if (ctx->clienthello_search) {
+		if (pxy_conn_autossl_peek_and_upgrade(ctx)) {
+			return;
+		}
+	}
+
 	struct evbuffer *inbuf = bufferevent_get_input(bev);
 	if (other->closed) {
 		evbuffer_drain(inbuf, evbuffer_get_length(inbuf));
@@ -1611,19 +1754,22 @@ pxy_bev_writecb(struct bufferevent *bev, void *arg)
 	}
 #endif /* DEBUG_PROXY */
 
-	struct evbuffer *outbuf = bufferevent_get_output(bev);
-	if (evbuffer_get_length(outbuf) > 0) {
+	if (other->closed) {
+		struct evbuffer *outbuf = bufferevent_get_output(bev);
+		if (evbuffer_get_length(outbuf) == 0) {
+			/* finished writing and other end is closed;
+			 * close this end too and clean up memory */
+			bufferevent_free_and_close_fd(bev, ctx);
+			pxy_conn_ctx_free(ctx);
+		}
+		return;
+	}
+
+	if (other->bev && !(bufferevent_get_enabled(other->bev) & EV_READ)) {
 		/* data source temporarily disabled;
 		 * re-enable and reset watermark to 0. */
 		bufferevent_setwatermark(bev, EV_WRITE, 0, 0);
-		if (!other->closed) {
-			bufferevent_enable(other->bev, EV_READ);
-		}
-	} else if (other->closed) {
-		/* finished writing and other end is closed;
-		 * close this end too and clean up memory */
-		bufferevent_free_and_close_fd(bev, ctx);
-		pxy_conn_ctx_free(ctx);
+		bufferevent_enable(other->bev, EV_READ);
 	}
 }
 
@@ -1664,7 +1810,8 @@ pxy_bev_eventcb(struct bufferevent *bev, short events, void *arg)
 		ctx->connected = 1;
 
 		/* wrap client-side socket in an eventbuffer */
-		if (ctx->spec->ssl && !ctx->passthrough) {
+		if ((ctx->spec->ssl || ctx->clienthello_found) &&
+		    !ctx->passthrough) {
 			ctx->src.ssl = pxy_srcssl_create(ctx, this->ssl);
 			if (!ctx->src.ssl) {
 				bufferevent_free_and_close_fd(bev, ctx);
@@ -1684,8 +1831,22 @@ pxy_bev_eventcb(struct bufferevent *bev, short events, void *arg)
 				return;
 			}
 		}
-		ctx->src.bev = pxy_bufferevent_setup(ctx, ctx->fd,
-		                                     ctx->src.ssl);
+		if (ctx->clienthello_found) {
+			if (OPTS_DEBUG(ctx->opts)) {
+				log_dbg_printf("Completing autossl upgrade\n");
+			}
+			ctx->src.bev = bufferevent_openssl_filter_new(
+			               ctx->evbase, ctx->src.bev, ctx->src.ssl,
+			               BUFFEREVENT_SSL_ACCEPTING,
+			               BEV_OPT_DEFER_CALLBACKS);
+			bufferevent_setcb(ctx->src.bev, pxy_bev_readcb,
+			                  pxy_bev_writecb, pxy_bev_eventcb,
+			                  ctx);
+			bufferevent_enable(ctx->src.bev, EV_READ|EV_WRITE);
+		} else {
+			ctx->src.bev = pxy_bufferevent_setup(ctx, ctx->fd,
+			                                     ctx->src.ssl);
+		}
 		if (!ctx->src.bev) {
 			if (ctx->src.ssl) {
 				SSL_free(ctx->src.ssl);
@@ -1699,10 +1860,10 @@ pxy_bev_eventcb(struct bufferevent *bev, short events, void *arg)
 
 		/* prepare logging, part 2 */
 		if (WANT_CONNECT_LOG(ctx) || WANT_CONTENT_LOG(ctx)) {
-			ctx->dst_str = sys_sockaddr_str((struct sockaddr *)
-			                                &ctx->addr,
-			                                ctx->addrlen);
-			if (!ctx->dst_str) {
+			if (sys_sockaddr_str((struct sockaddr *)
+			                     &ctx->addr, ctx->addrlen,
+			                     &ctx->dsthost_str,
+			                     &ctx->dstport_str) != 0) {
 				ctx->enomem = 1;
 				pxy_conn_terminate_free(ctx);
 				return;
@@ -1736,7 +1897,8 @@ pxy_bev_eventcb(struct bufferevent *bev, short events, void *arg)
 		}
 		if (WANT_CONTENT_LOG(ctx)) {
 			if (log_content_open(&ctx->logctx, ctx->opts,
-			                     ctx->src_str, ctx->dst_str,
+			                     ctx->srchost_str, ctx->srcport_str,
+			                     ctx->dsthost_str, ctx->dstport_str,
 #ifdef HAVE_LOCAL_PROCINFO
 			                     ctx->lproc.exec_path,
 			                     ctx->lproc.user,
@@ -1760,14 +1922,25 @@ connected:
 			pxy_log_connect_nonhttp(ctx);
 		}
 
+		/* write SSL certificates to gendir */
+		if (this->ssl && (bev == ctx->src.bev) &&
+		    ctx->opts->certgendir) {
+			pxy_srccert_write(ctx);
+		}
+
 		if (OPTS_DEBUG(ctx->opts)) {
 			if (this->ssl) {
 				/* for SSL, we get two connect events */
-				log_dbg_printf("SSL connected %s %s %s %s\n",
+				log_dbg_printf("SSL connected %s [%s]:%s"
+				               " %s %s\n",
 				               bev == ctx->dst.bev ?
 				               "to" : "from",
 				               bev == ctx->dst.bev ?
-				               ctx->dst_str : ctx->src_str,
+				               ctx->dsthost_str :
+				               ctx->srchost_str,
+				               bev == ctx->dst.bev ?
+				               ctx->dstport_str :
+				               ctx->srcport_str,
 				               SSL_get_version(this->ssl),
 				               SSL_get_cipher(this->ssl));
 			} else {
@@ -1776,10 +1949,12 @@ connected:
 				 * beginning; mirror SSL debug output anyway
 				 * in order not to confuse anyone who might be
 				 * looking closely at the output */
-				log_dbg_printf("TCP connected to %s\n",
-				               ctx->dst_str);
-				log_dbg_printf("TCP connected from %s\n",
-				               ctx->src_str);
+				log_dbg_printf("TCP connected to [%s]:%s\n",
+				               ctx->dsthost_str,
+				               ctx->dstport_str);
+				log_dbg_printf("TCP connected from [%s]:%s\n",
+				               ctx->srchost_str,
+				               ctx->srcport_str);
 			}
 		}
 
@@ -1892,6 +2067,7 @@ connected:
 			outbuf = bufferevent_get_output(other->bev);
 			if (evbuffer_get_length(outbuf) == 0) {
 				bufferevent_free_and_close_fd(other->bev, ctx);
+				other->bev = NULL;
 				other->closed = 1;
 			}
 		}
@@ -1899,7 +2075,12 @@ connected:
 	}
 
 	if (events & BEV_EVENT_EOF) {
-		if (!other->closed) {
+		if (!ctx->connected) {
+			log_dbg_printf("EOF on inbound connection while "
+			               "connecting to original destination\n");
+			evutil_closesocket(ctx->fd);
+			other->closed = 1;
+		} else if (!other->closed) {
 			struct evbuffer *inbuf, *outbuf;
 			inbuf = bufferevent_get_input(bev);
 			outbuf = bufferevent_get_output(other->bev);
@@ -1913,6 +2094,7 @@ connected:
 				if (evbuffer_get_length(outbuf) == 0) {
 					bufferevent_free_and_close_fd(
 							other->bev, ctx);
+					other->bev = NULL;
 					other->closed = 1;
 				}
 			}
@@ -1926,16 +2108,17 @@ connected:
 leave:
 	/* we only get a single disconnect event here for both connections */
 	if (OPTS_DEBUG(ctx->opts)) {
-		log_dbg_printf("%s disconnected to %s\n",
+		log_dbg_printf("%s disconnected to [%s]:%s\n",
 		               this->ssl ? "SSL" : "TCP",
-		               ctx->dst_str);
-		log_dbg_printf("%s disconnected from %s\n",
+		               ctx->dsthost_str, ctx->dstport_str);
+		log_dbg_printf("%s disconnected from [%s]:%s\n",
 		               this->ssl ? "SSL" : "TCP",
-		               ctx->src_str);
+		               ctx->srchost_str, ctx->srcport_str);
 	}
 
 	this->closed = 1;
 	bufferevent_free_and_close_fd(bev, ctx);
+	this->bev = NULL;
 	if (other->closed) {
 		pxy_conn_ctx_free(ctx);
 	}
@@ -1977,11 +2160,15 @@ pxy_conn_connect(pxy_conn_ctx_t *ctx)
 	}
 
 	if (OPTS_DEBUG(ctx->opts)) {
-		char *ip = sys_sockaddr_str((struct sockaddr *)&ctx->addr,
-		                            ctx->addrlen);
-		log_dbg_printf("Connecting to %s\n", ip);
-		if (ip)
-			free(ip);
+		char *host, *port;
+		if (sys_sockaddr_str((struct sockaddr *)&ctx->addr,
+		                     ctx->addrlen, &host, &port) != 0) {
+			log_dbg_printf("Connecting to [?]:?\n");
+		} else {
+			log_dbg_printf("Connecting to [%s]:%s\n", host, port);
+			free(host);
+			free(port);
+		}
 	}
 
 	/* initiate connection */
@@ -2033,10 +2220,12 @@ pxy_fd_readcb(MAYBE_UNUSED evutil_socket_t fd, UNUSED short what, void *arg)
 	pxy_conn_ctx_t *ctx = arg;
 
 #ifndef OPENSSL_NO_TLSEXT
-	/* for SSL, peek clientHello and parse SNI from it */
+	/* for SSL, peek ClientHello and parse SNI from it */
 	if (ctx->spec->ssl && !ctx->passthrough /*&& ctx->ev*/) {
 		unsigned char buf[1024];
 		ssize_t n;
+		const unsigned char *chello;
+		int rv;
 
 		n = recv(fd, buf, sizeof(buf), MSG_PEEK);
 		if (n == -1) {
@@ -2053,15 +2242,23 @@ pxy_fd_readcb(MAYBE_UNUSED evutil_socket_t fd, UNUSED short what, void *arg)
 			return;
 		}
 
-		ctx->sni = ssl_tls_clienthello_parse_sni(buf, &n);
+		rv = ssl_tls_clienthello_parse(buf, n, 0, &chello, &ctx->sni);
+		if ((rv == 1) && !chello) {
+			log_err_printf("Peeking did not yield a (truncated) "
+			               "ClientHello message, "
+			               "aborting connection\n");
+			evutil_closesocket(fd);
+			pxy_conn_ctx_free(ctx);
+			return;
+		}
 		if (OPTS_DEBUG(ctx->opts)) {
 			log_dbg_printf("SNI peek: [%s] [%s]\n",
 			               ctx->sni ? ctx->sni : "n/a",
-			               (!ctx->sni && (n == -1)) ?
+			               ((rv == 1) && chello) ?
 			               "incomplete" : "complete");
 		}
-		if (!ctx->sni && (n == -1) && (ctx->sni_peek_retries++ < 50)) {
-			/* ssl_tls_clienthello_parse_sni indicates that we
+		if ((rv == 1) && chello && (ctx->sni_peek_retries++ < 50)) {
+			/* ssl_tls_clienthello_parse indicates that we
 			 * should retry later when we have more data, and we
 			 * haven't reached the maximum retry count yet.
 			 * Reschedule this event as timeout-only event in
@@ -2168,8 +2365,9 @@ pxy_conn_setup(evutil_socket_t fd,
 
 	/* prepare logging, part 1 */
 	if (WANT_CONNECT_LOG(ctx) || WANT_CONTENT_LOG(ctx)) {
-		ctx->src_str = sys_sockaddr_str(peeraddr, peeraddrlen);
-		if (!ctx->src_str)
+		if (sys_sockaddr_str(peeraddr, peeraddrlen,
+		                     &ctx->srchost_str,
+		                     &ctx->srcport_str) != 0)
 			goto memout;
 #ifdef HAVE_LOCAL_PROCINFO
 		if (ctx->opts->lprocinfo) {
